//! # Dispatcher
//!
//! Here is where every command is defined. Each command has some definition and a handler. Their
//! handler are rust functions.
//!
//! Each command is defined with the dispatcher macro, which generates efficient and developer
//! friendly code.
use crate::{
    cmd,
    connection::{Connection, ConnectionStatus},
    dispatcher,
    error::Error,
    value::Value,
};
use bytes::Bytes;
use std::convert::TryInto;
use std::time::SystemTime;
use std::time::UNIX_EPOCH;

pub mod command;

/// Returns the server time
async fn do_time(_conn: &Connection, _args: &[Bytes]) -> Result<Value, Error> {
    let now = SystemTime::now();
    let since_the_epoch = now.duration_since(UNIX_EPOCH).expect("Time went backwards");
    let seconds = format!("{}", since_the_epoch.as_secs());
    let millis = format!("{}", since_the_epoch.subsec_millis());

    Ok(vec![seconds.as_str(), millis.as_str()].into())
}

dispatcher! {
    set {
        sadd {
            cmd::set::sadd,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        scard {
            cmd::set::scard,
            [""],
            2,
            1,
            1,
            1,
            true,
        },
        sdiff {
            cmd::set::sdiff,
            [""],
            -2,
            1,
            -1,
            1,
            true,
        },
        sdiffstore {
            cmd::set::sdiffstore,
            [""],
            -3,
            1,
            -1,
            1,
            true,
        },
        sinter {
            cmd::set::sinter,
            [""],
            -2,
            1,
            -1,
            1,
            true,
        },
        sintercard {
            cmd::set::sintercard,
            [""],
            -2,
            1,
            -1,
            1,
            true,
        },
        sinterstore {
            cmd::set::sinterstore,
            [""],
            -3,
            1,
            -1,
            1,
            true,
        },
        sismember {
            cmd::set::sismember,
            [""],
            3,
            1,
            1,
            1,
            true,
        },
        smembers {
            cmd::set::smembers,
            [""],
            2,
            1,
            1,
            1,
            true,
        },
        smismember {
            cmd::set::smismember,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        smove {
            cmd::set::smove,
            [""],
            4,
            1,
            2,
            1,
            true,
        },
        spop {
            cmd::set::spop,
            [""],
            -2,
            1,
            1,
            1,
            true,
        },
        srandmember {
            cmd::set::srandmember,
            [""],
            -2,
            1,
            1,
            1,
            true,
        },
        srem {
            cmd::set::srem,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        sunion {
            cmd::set::sunion,
            [""],
            -2,
            1,
            -1,
            1,
            true,
        },
        sunionstore {
            cmd::set::sunionstore,
            [""],
            -2,
            1,
            -1,
            1,
            true,
        },
    },
    metrics {
        metrics {
            cmd::metrics::metrics,
            [""],
            -1,
            0,
            0,
            0,
            false,
        },
    },
    list {
        blpop {
            cmd::list::blpop,
            [""],
            -3,
            1,
            -2,
            1,
            true,
        },
        brpop {
            cmd::list::brpop,
            [""],
            -3,
            1,
            -2,
            1,
            true,
        },
        lindex {
            cmd::list::lindex,
            [""],
            3,
            1,
            1,
            1,
            true,
        },
        linsert {
            cmd::list::linsert,
            [""],
            5,
            1,
            1,
            1,
            true,
        },
        llen {
            cmd::list::llen,
            [""],
            2,
            1,
            1,
            1,
            true,
        },
        lmove {
            cmd::list::lmove,
            [""],
            5,
            1,
            2,
            1,
            true,
        },
        lpop {
            cmd::list::lpop,
            [""],
            -2,
            1,
            -2,
            1,
            true,
        },
        lpos {
            cmd::list::lpos,
            [""],
            -2,
            1,
            1,
            1,
            true,
        },
        lpush {
            cmd::list::lpush,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        lpushx {
            cmd::list::lpush,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        lrange {
            cmd::list::lrange,
            [""],
            4,
            1,
            1,
            1,
            true,
        },
        lrem {
            cmd::list::lrem,
            [""],
            4,
            1,
            1,
            1,
            true,
        },
        lset {
            cmd::list::lset,
            [""],
            4,
            1,
            1,
            1,
            true,
        },
        ltrim {
            cmd::list::ltrim,
            [""],
            4,
            1,
            1,
            1,
            true,
        },
        rpop {
            cmd::list::rpop,
            [""],
            -2,
            1,
            1,
            1,
            true,
        },
        rpoplpush {
            cmd::list::rpoplpush,
            [""],
            3,
            1,
            2,
            1,
            true,
        },
        rpush {
            cmd::list::rpush,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        rpushx {
            cmd::list::rpush,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
    },
    hash {
        hdel {
            cmd::hash::hdel,
            [""],
            -2,
            1,
            1,
            1,
            true,
        },
        hexists {
            cmd::hash::hexists,
            [""],
            3,
            1,
            1,
            1,
            true,
        },
        hget {
            cmd::hash::hget,
            [""],
            3,
            1,
            1,
            1,
            true,
        },
        hgetall {
            cmd::hash::hgetall,
            [""],
            2,
            1,
            1,
            1,
            true,
        },
        hincrby {
            cmd::hash::hincrby::<i64>,
            [""],
            4,
            1,
            1,
            1,
            true,
        },
        hincrbyfloat {
            cmd::hash::hincrby::<f64>,
            [""],
            4,
            1,
            1,
            1,
            true,
        },
        hkeys {
            cmd::hash::hkeys,
            [""],
            2,
            1,
            1,
            1,
            true,
        },
        hlen {
            cmd::hash::hlen,
            [""],
            2,
            1,
            1,
            1,
            true,
        },
        hmget {
            cmd::hash::hmget,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        hmset {
            cmd::hash::hset,
            [""],
            -3,
            1,
            1,
            1,
            true,
        },
        hrandfield {
            cmd::hash::hrandfield,
            [""],
            -2,
            1,
            1,
            1,
            true,
        },
        hset {
            cmd::hash::hset,
            [""],
            -4,
            1,
            1,
            1,
            true,
        },
        hsetnx {
            cmd::hash::hsetnx,
            [""],
            4,
            1,
            1,
            1,
            true,
        },
        hstrlen {
            cmd::hash::hstrlen,
            [""],
            3,
            1,
            1,
            1,
            true,
        },
        hvals {
            cmd::hash::hvals,
            [""],
            2,
            1,
            1,
            1,
            true,
        },
    },
    keys {
        del {
            cmd::key::del,
            ["random" "loading" "stale"],
            -2,
            1,
            -1,
            1,
            true,
        },
        exists {
            cmd::key::exists,
            ["read" "fast"],
            -2,
            1,
            -1,
            1,
            true,
        },
        expire {
            cmd::key::expire,
            ["read" "write" "fast"],
            3,
            1,
            1,
            1,
            true,
        },
        expireat {
            cmd::key::expire_at,
            ["read" "write" "fast"],
            3,
            1,
            1,
            1,
            true,
        },
        expiretime {
            cmd::key::expire_time,
            ["read" "write" "fast"],
            2,
            1,
            1,
            1,
            true,
        },
        persist {
            cmd::key::persist,
            ["write" "fast"],
            2,
            1,
            1,
            1,
            true,
        },
        pexpire {
            cmd::key::expire,
            ["read" "write" "fast"],
            3,
            1,
            1,
            1,
            true,
        },
        pexpireat {
            cmd::key::expire_at,
            ["read" "write" "fast"],
            3,
            1,
            1,
            1,
            true,
        },
        pexpiretime {
            cmd::key::expire_time,
            ["read" "write" "fast"],
            2,
            1,
            1,
            1,
            true,
        },
        pttl {
            cmd::key::ttl,
            ["read" "read"],
            2,
            1,
            1,
            1,
            true,
        },
        ttl {
            cmd::key::ttl,
            ["read" "read"],
            2,
            1,
            1,
            1,
            true,
        },
    },
    string {
        decr {
            cmd::string::decr,
            ["write" "denyoom" "fast"],
            2,
            1,
            1,
            1,
            true,
        },
        decrby {
            cmd::string::decr_by,
            ["write" "denyoom" "fast"],
            3,
            1,
            1,
            1,
            true,
        },
        get {
            cmd::string::get,
            ["random" "loading" "stale"],
            2,
            1,
            1,
            1,
            true,
        },
        getdel {
            cmd::string::getdel,
            ["random" "loading" "stale"],
            2,
            1,
            1,
            1,
            true,
        },
        getset {
            cmd::string::getset,
            ["random" "loading" "stale"],
            3,
            1,
            1,
            1,
            true,
        },
        incr {
            cmd::string::incr,
            ["write" "denyoom" "fast"],
            2,
            1,
            1,
            1,
            true,
        },
        incrby {
            cmd::string::incr_by,
            ["write" "denyoom" "fast"],
            3,
            1,
            1,
            1,
            true,
        },
        incrbyfloat {
            cmd::string::incr_by_float,
            ["write" "denyoom" "fast"],
            3,
            1,
            1,
            1,
            true,
        },
        mget {
            cmd::string::mget,
            ["random" "loading" "stale"],
            -2,
            1,
            -1,
            1,
            true,
        },
        set {
            cmd::string::set,
            ["random" "loading" "stale"],
            -3,
            1,
            1,
            1,
            true,
        },
        setex {
            cmd::string::setex,
            ["random" "loading" "stale"],
            4,
            1,
            1,
            1,
            true,
        },
        psetex {
            cmd::string::setex,
            ["random" "loading" "stale"],
            4,
            1,
            1,
            1,
            true,
        },
        strlen {
            cmd::string::strlen,
            ["random" "fast"],
            2,
            1,
            1,
            1,
            true,
        }
    },
    connection {
        client {
            cmd::client::client,
            ["random" "loading" "stale"],
            -2,
            0,
            0,
            0,
            true,
        },
        echo {
            cmd::client::echo,
            ["random" "loading" "stale"],
            2,
            0,
            0,
            0,
            true,
        },
        ping {
            cmd::client::ping,
            ["random" "loading" "stale"],
            -1,
            0,
            0,
            0,
            true,
        },
        reset {
            cmd::client::reset,
            [""],
            1,
            0,
            0,
            0,
            false,
        },
    },
    transaction {
        discard {
            cmd::transaction::discard,
            [""],
            1,
            0,
            0,
            0,
            false,
        },
        exec {
            cmd::transaction::exec,
            [""],
            1,
            0,
            0,
            0,
            false,
        },
        multi {
            cmd::transaction::multi,
            [""],
            1,
            0,
            0,
            0,
            false,
        },
        watch {
            cmd::transaction::watch,
            [""],
            -2,
            1,
            -1,
            1,
            false,
        },
        unwatch {
            cmd::transaction::unwatch,
            [""],
            1,
            0,
            0,
            0,
            true,
        },
    },
    pubsub {
        publish {
            cmd::pubsub::publish,
            [""],
            3,
            0,
            0,
            0,
            true,
        },
        pubsub {
            cmd::pubsub::pubsub,
            [""],
            -2,
            0,
            0,
            0,
            true,
        },
        psubscribe {
            cmd::pubsub::subscribe,
            [""],
            -2,
            0,
            0,
            0,
            true,
        },
        punsubscribe {
            cmd::pubsub::punsubscribe,
            [""],
            -1,
            0,
            0,
            0,
            true,
        },
        subscribe {
            cmd::pubsub::subscribe,
            [""],
            -2,
            0,
            0,
            0,
            true,
        },
        unsubscribe {
            cmd::pubsub::unsubscribe,
            [""],
            -1,
            0,
            0,
            0,
            true,
        },
    },
    server {
        time {
            do_time,
            ["random" "loading" "stale"],
            1,
            0,
            0,
            0,
            true,
        },
    }
}
